@@ -7,12 +7,16 @@
 //!
 
 use std::{
-    borrow::Borrow,
+    borrow::{Borrow, Cow},
     collections::HashSet,
+    ffi::{OsStr, OsString},
     fs::{self, File, OpenOptions},
     io::{BufRead, BufReader, Write},
-    os::unix::{ffi::OsStrExt, fs::PermissionsExt},
-    path::{Path, PathBuf},
+    os::unix::{
+        ffi::{OsStrExt, OsStringExt},
+        fs::PermissionsExt,
+    },
+    path::{Component, Path, PathBuf},
 };
 
 use log::{debug, warn};
@@ -44,6 +48,9 @@ impl TrashContext {
                 // Note that the following function creates the trash folder
                 // and its required subfolders in case they don't exist.
                 move_to_trash(path, &home_trash, topdir).map_err(|(p, e)| fs_error(p, e))?;
+            } else if topdir.to_str() == Some("/var/home") && home_topdir.to_str() == Some("/") {
+                debug!("The topdir is '/var/home' but the home_topdir is '/', moving to the home trash anyway.");
+                move_to_trash(path, &home_trash, topdir).map_err(|(p, e)| fs_error(p, e))?;
             } else {
                 execute_on_mounted_trash_folders(uid, topdir, true, true, |trash_path| {
                     move_to_trash(&path, trash_path, topdir)
@@ -56,39 +63,8 @@ impl TrashContext {
 }
 
 pub fn list() -> Result<Vec<TrashItem>, Error> {
-    let mut trash_folders = HashSet::new();
-    // Get home trash folder and add it to the set of trash folders.
-    // It may not exist and that's completely fine as long as there are other trash folders.
-    let home_error;
-    match home_trash() {
-        Ok(home_trash) => {
-            if !home_trash.is_dir() {
-                home_error = Some(Error::Unknown {
-                    description:
-                        "The 'home trash' either does not exist or is not a directory (or a link pointing to a dir)"
-                            .into(),
-                });
-            } else {
-                trash_folders.insert(home_trash);
-                home_error = None;
-            }
-        }
-        Err(e) => {
-            home_error = Some(e);
-        }
-    }
+    let EvaluatedTrashFolders { trash_folders, home_error, sorted_mount_points } = eval_trash_folders()?;
 
-    // Get all mount-points and attempt to find a trash folder in each adding them to the SET of
-    // trash folders when found one.
-    let uid = unsafe { libc::getuid() };
-    let sorted_mount_points = get_sorted_mount_points()?;
-    for mount in &sorted_mount_points {
-        execute_on_mounted_trash_folders(uid, &mount.mnt_dir, false, false, |trash_path| {
-            trash_folders.insert(trash_path);
-            Ok(())
-        })
-        .map_err(|(p, e)| fs_error(p, e))?;
-    }
     if trash_folders.is_empty() {
         warn!("No trash folder was found. The error when looking for the 'home trash' was: {:?}", home_error);
         return Ok(vec![]);
@@ -167,13 +143,16 @@ pub fn list() -> Result<Vec<TrashItem>, Error> {
                 let value = split.next().unwrap().trim();
 
                 if key == "Path" {
-                    let mut value_path = Path::new(value).to_owned();
-                    if value_path.is_relative() {
-                        value_path = top_dir.join(value_path);
-                    }
-                    let full_path_utf8 = PathBuf::from(parse_uri_path(&value_path));
-                    name = Some(full_path_utf8.file_name().unwrap().to_str().unwrap().to_owned());
-                    let parent = full_path_utf8.parent().unwrap();
+                    let value_path = {
+                        let path = Path::new(value);
+                        if path.is_relative() {
+                            decode_uri_path(top_dir.join(path))
+                        } else {
+                            decode_uri_path(path)
+                        }
+                    };
+                    name = value_path.file_name().map(|name| name.to_owned());
+                    let parent = value_path.parent().expect("Absolute path to trashed item should have a parent");
                     original_parent = Some(parent.into());
                 } else if key == "DeletionDate" {
                     #[cfg(feature = "chrono")]
@@ -218,6 +197,90 @@ pub fn list() -> Result<Vec<TrashItem>, Error> {
     Ok(result)
 }
 
+pub fn is_empty() -> Result<bool, Error> {
+    let trash_folders = trash_folders()?;
+
+    if trash_folders.is_empty() {
+        return Ok(true);
+    }
+
+    for folder in trash_folders {
+        // We're only concerned if the trash contains any files
+        // Therefore, we only need to check if the bin itself is empty
+        let bin = folder.join("files");
+        match bin.read_dir() {
+            Ok(mut entries) => {
+                if let Some(Ok(_)) = entries.next() {
+                    return Ok(false);
+                }
+            }
+            Err(e) => {
+                warn!("The trash files folder {:?} could not be read. Error was {:?}", bin, e);
+            }
+        }
+    }
+
+    Ok(true)
+}
+
+pub fn trash_folders() -> Result<HashSet<PathBuf>, Error> {
+    let EvaluatedTrashFolders { trash_folders, home_error, .. } = eval_trash_folders()?;
+
+    if trash_folders.is_empty() {
+        return match home_error {
+            Some(e) => Err(e),
+            None => Err(Error::Unknown {
+                description: "Could not find a valid 'home trash' nor valid trashes on other mount points".into(),
+            }),
+        };
+    }
+
+    Ok(trash_folders)
+}
+
+struct EvaluatedTrashFolders {
+    trash_folders: HashSet<PathBuf>,
+    home_error: Option<Error>,
+    sorted_mount_points: Vec<MountPoint>,
+}
+
+fn eval_trash_folders() -> Result<EvaluatedTrashFolders, Error> {
+    let mut trash_folders = HashSet::new();
+    // Get home trash folder and add it to the set of trash folders.
+    // It may not exist and that's completely fine as long as there are other trash folders.
+    let home_error;
+    match home_trash() {
+        Ok(home_trash) => {
+            if !home_trash.is_dir() {
+                home_error = Some(Error::Unknown {
+                    description:
+                        "The 'home trash' either does not exist or is not a directory (or a link pointing to a dir)"
+                            .into(),
+                });
+            } else {
+                trash_folders.insert(home_trash);
+                home_error = None;
+            }
+        }
+        Err(e) => {
+            home_error = Some(e);
+        }
+    }
+
+    // Get all mount-points and attempt to find a trash folder in each adding them to the SET of
+    // trash folders when found one.
+    let uid = unsafe { libc::getuid() };
+    let sorted_mount_points = get_sorted_mount_points()?;
+    for mount in &sorted_mount_points {
+        execute_on_mounted_trash_folders(uid, &mount.mnt_dir, false, false, |trash_path| {
+            trash_folders.insert(trash_path);
+            Ok(())
+        })
+        .map_err(|(p, e)| fs_error(p, e))?;
+    }
+
+    Ok(EvaluatedTrashFolders { trash_folders, home_error, sorted_mount_points })
+}
 pub fn metadata(item: &TrashItem) -> Result<TrashItemMetadata, Error> {
     // When purging an item the "in-trash" filename must be parsed from the trashinfo filename
     // which is the filename in the `id` field.
@@ -412,15 +475,20 @@ fn move_to_trash(
     // already exist. This newly created empty file can then be safely overwritten by the src file
     // using the `rename` function.
     let filename = src.file_name().unwrap();
-    let mut appendage = 0;
+    let mut appendage = 0usize;
     loop {
         appendage += 1;
-        let in_trash_name = if appendage > 1 {
-            format!("{}.{}", filename.to_str().unwrap(), appendage)
+        let in_trash_name: Cow<'_, OsStr> = if appendage > 1 {
+            let mut trash_name = filename.to_owned();
+            trash_name.push(format!(".{appendage}"));
+            trash_name.into()
         } else {
-            filename.to_str().unwrap().into()
+            filename.into()
         };
-        let info_name = format!("{in_trash_name}.trashinfo");
+        // Length of name + length of '.trashinfo'
+        let mut info_name = OsString::with_capacity(in_trash_name.len() + 10);
+        info_name.push(&in_trash_name);
+        info_name.push(".trashinfo");
         let info_file_path = info_folder.join(&info_name);
         let info_result = OpenOptions::new().create_new(true).write(true).open(&info_file_path);
         match info_result {
@@ -477,54 +545,13 @@ fn move_to_trash(
     Ok(())
 }
 
-fn execute_src_to_dst_operation<S1, D1>(
-    src: S1,
-    dst: D1,
-    dir: &'static dyn Fn(&Path) -> Result<(), FsError>,
-    file: &'static dyn Fn(&Path, &Path) -> Result<(), FsError>,
-) -> Result<(), FsError>
-where
-    S1: AsRef<Path>,
-    D1: AsRef<Path>,
-{
-    let src = src.as_ref();
-    let dst = dst.as_ref();
-
-    let metadata = src.symlink_metadata().map_err(|e| (src.to_owned(), e))?;
-    if metadata.is_dir() {
-        dir(dst)?;
-        let dir_entries = std::fs::read_dir(src).map_err(|e| (src.to_owned(), e))?;
-        for entry in dir_entries {
-            // Forward the error because it's not okay if something is happening
-            // to the files while we are trying to move them.
-            let entry = entry.map_err(|e| (src.to_owned(), e))?;
-            let entry_src = entry.path();
-            let entry_dst = dst.join(entry.file_name());
-            execute_src_to_dst_operation(entry_src, entry_dst, dir, file)?;
-        }
-    } else {
-        // Symlink or file
-        file(src, dst)?;
-    }
-    Ok(())
-}
-
 /// An error may mean that a collision was found.
 fn move_items_no_replace(src: impl AsRef<Path>, dst: impl AsRef<Path>) -> Result<(), FsError> {
     let src = src.as_ref();
     let dst = dst.as_ref();
 
     try_creating_placeholders(src, dst)?;
-
-    // All placeholders are in place. LET'S OVERWRITE
-    execute_src_to_dst_operation(src, dst, &|_| Ok(()), &|src, dst| {
-        if let Some(parent) = dst.parent() {
-            if let Err(err) = std::fs::create_dir_all(parent) {
-                warn!("Failed to create destination directory. It probably already exists. {:?}", err);
-            }
-        }
-        std::fs::rename(src, dst).map_err(|e| (src.to_owned(), e))
-    })?;
+    std::fs::rename(src, dst).map_err(|e| (src.to_owned(), e))?;
 
     // Once everything is moved, lets recursively remove the directory
     if src.is_dir() {
@@ -547,15 +574,31 @@ fn try_creating_placeholders(src: impl AsRef<Path>, dst: impl AsRef<Path>) -> Re
     Ok(())
 }
 
-fn parse_uri_path(absolute_file_path: impl AsRef<Path>) -> String {
-    let file_path_chars = absolute_file_path.as_ref().to_str().unwrap().chars();
-    let url: String = "file://".chars().chain(file_path_chars).collect();
-    return url::Url::parse(&url).unwrap().to_file_path().unwrap().to_str().unwrap().into();
+fn decode_uri_path(path: impl AsRef<Path>) -> PathBuf {
+    // Paths may be invalid Unicode on most Unixes so they should be treated as byte strings
+    // A higher level crate, such as `url`, can't be used directly since its API intakes valid Rust
+    // strings. Thus, the easiest way is to manually decode each segment of the path and recombine.
+    path.as_ref().iter().map(|part| OsString::from_vec(urlencoding::decode_binary(part.as_bytes()).to_vec())).collect()
 }
 
-fn encode_uri_path(absolute_file_path: impl AsRef<Path>) -> String {
-    let url = url::Url::from_file_path(absolute_file_path.as_ref()).unwrap();
-    url.path().to_owned()
+fn encode_uri_path(path: impl AsRef<Path>) -> String {
+    // `iter()` cannot be used here because it yields '/' in certain situations, such as
+    // for root directories.
+    // Slashes would be encoded and thus mess up the path
+    let path: PathBuf = path
+        .as_ref()
+        .components()
+        .map(|component| {
+            // Only encode names and not '/', 'C:\', et cetera
+            if let Component::Normal(part) = component {
+                urlencoding::encode_binary(part.as_bytes()).to_string()
+            } else {
+                component.as_os_str().to_str().expect("Path components such as '/' are valid Unicode").to_owned()
+            }
+        })
+        .collect();
+
+    path.to_str().expect("URL encoded bytes is valid Unicode").to_owned()
 }
 
 #[derive(Eq, PartialEq, Debug)]
@@ -822,10 +865,10 @@ mod tests {
     use std::{
         collections::{hash_map::Entry, HashMap},
         env,
-        ffi::OsString,
+        ffi::{OsStr, OsString},
         fmt,
         fs::File,
-        os::unix,
+        os::unix::{self, ffi::OsStringExt},
         path::{Path, PathBuf},
         process::Command,
     };
@@ -835,10 +878,13 @@ mod tests {
     use crate::{
         canonicalize_paths, delete, delete_all,
         os_limited::{list, purge_all, restore_all},
+        platform::encode_uri_path,
         tests::get_unique_name,
         Error,
     };
 
+    use super::decode_uri_path;
+
     #[test]
     #[serial]
     fn test_list() {
@@ -847,7 +893,7 @@ mod tests {
         let file_name_prefix = get_unique_name();
         let batches: usize = 2;
         let files_per_batch: usize = 3;
-        let names: Vec<_> = (0..files_per_batch).map(|i| format!("{}#{}", file_name_prefix, i)).collect();
+        let names: Vec<OsString> = (0..files_per_batch).map(|i| format!("{}#{}", file_name_prefix, i).into()).collect();
         for _ in 0..batches {
             for path in &names {
                 File::create(path).unwrap();
@@ -863,8 +909,10 @@ mod tests {
             }
         }
         let items = list().unwrap();
-        let items: HashMap<_, Vec<_>> =
-            items.into_iter().filter(|x| x.name.starts_with(&file_name_prefix)).fold(HashMap::new(), |mut map, x| {
+        let items: HashMap<_, Vec<_>> = items
+            .into_iter()
+            .filter(|x| x.name.as_encoded_bytes().starts_with(file_name_prefix.as_bytes()))
+            .fold(HashMap::new(), |mut map, x| {
                 match map.entry(x.name.clone()) {
                     Entry::Occupied(mut entry) => {
                         entry.get_mut().push(x);
@@ -878,7 +926,7 @@ mod tests {
         for name in names {
             match items.get(&name) {
                 Some(items) => assert_eq!(items.len(), batches),
-                None => panic!("ERROR Could not find '{}' in {:#?}", name, items),
+                None => panic!("ERROR Could not find '{:?}' in {:#?}", name, items),
             }
         }
 
@@ -894,8 +942,8 @@ mod tests {
 
         let file_name_prefix = get_unique_name();
         let file_count = 2;
-        let names: Vec<_> = (0..file_count).map(|i| format!("{}#{}", file_name_prefix, i)).collect();
-        let symlink_names: Vec<_> = names.iter().map(|name| format!("{}-symlink", name)).collect();
+        let names: Vec<OsString> = (0..file_count).map(|i| format!("{}#{}", file_name_prefix, i).into()).collect();
+        let symlink_names: Vec<OsString> = names.iter().map(|name| format!("{:?}-symlink", name).into()).collect();
 
         // Test file symbolic link and directory symbolic link
         File::create(&names[0]).unwrap();
@@ -924,6 +972,34 @@ mod tests {
         }
     }
 
+    #[test]
+    fn uri_enc_dec_roundtrip() {
+        let fake = format!("/tmp/{}", get_unique_name());
+        let path = decode_uri_path(&fake);
+
+        assert_eq!(path.to_str().expect("Path is valid Unicode"), fake, "Decoded path shouldn't be different");
+
+        let encoded = encode_uri_path(&path);
+        assert_eq!(encoded, fake, "URL encoded alphanumeric String shouldn't change");
+    }
+
+    #[test]
+    fn uri_enc_dec_roundtrip_invalid_unicode() {
+        let base = OsStr::new(&format!("/tmp/{}/", get_unique_name())).to_os_string();
+
+        // Add invalid UTF-8 byte
+        let mut bytes = base.into_encoded_bytes();
+        bytes.push(168);
+        let fake = OsString::from_vec(bytes);
+        assert!(fake.to_str().is_none(), "Invalid Unicode cannot be a Rust String");
+
+        let path = decode_uri_path(&fake);
+        assert_eq!(path.as_os_str().as_encoded_bytes(), fake.as_encoded_bytes());
+
+        // Shouldn't panic
+        encode_uri_path(&path);
+    }
+
     //////////////////////////////////////////////////////////////////////////////////////
     /// System
     //////////////////////////////////////////////////////////////////////////////////////
