@@ -274,7 +274,7 @@ pub struct TrashItem {
 
     /// The name of the item. For example if the folder '/home/user/New Folder'
     /// was deleted, its `name` is 'New Folder'
-    pub name: String,
+    pub name: OsString,
 
     /// The path to the parent folder of this item before it was put inside the
     /// trash. For example if the folder '/home/user/New Folder' is in the
@@ -374,6 +374,44 @@ pub mod os_limited {
         platform::list()
     }
 
+    /// Returns whether the trash is empty or has at least one item.
+    ///
+    /// Unlike calling [`list`], this function short circuits without evaluating every item.
+    ///
+    /// # Example
+    ///
+    /// ```
+    /// use trash::os_limited::is_empty;
+    /// if is_empty().unwrap_or(true) {
+    ///     println!("Trash is empty");
+    /// } else {
+    ///     println!("Trash contains at least one item");
+    /// }
+    /// ```
+    pub fn is_empty() -> Result<bool, Error> {
+        platform::is_empty()
+    }
+
+    /// Returns all valid trash bins on supported Unix platforms.
+    ///
+    /// Valid trash folders include the user's personal "home trash" as well as designated trash
+    /// bins across mount points. Some, or all of these, may not exist or be invalid in some way.
+    ///
+    /// # Example
+    ///
+    /// ```
+    /// # #[cfg(all(unix, not(target_os = "macos"), not(target_os = "ios"), not(target_os = "android")))] {
+    /// use trash::os_limited::trash_folders;
+    /// let trash_bins = trash_folders()?;
+    /// println!("{trash_bins:#?}");
+    /// # }
+    /// # Ok::<(), trash::Error>(())
+    /// ```
+    #[cfg(all(unix, not(target_os = "macos"), not(target_os = "ios"), not(target_os = "android")))]
+    pub fn trash_folders() -> Result<HashSet<std::path::PathBuf>, Error> {
+        platform::trash_folders()
+    }
+
     /// Returns the [`TrashItemMetadata`] for a [`TrashItem`]
     ///
     /// # Example
